//
//  Copyright (c) 1997 by Mark F. Adams 
//  $Id: prom_elem.C,v 1.3 2011-01-22 20:13:39 madams Exp $
//  Author:             Mark F. Adams
//  Version:            1.0
//  Creation Date:      23rd June, 1998
//  Filename:           element.C
//
#include "prom_elem.hh"

// temp elements used in FEAP and in 'fill'
const int PromPoly::tetID_[4][4]={ {3,2,1,-9},{3,0,2,-9},
				   {3,1,0,-9},{0,1,2,-9}};
// temp used in insertion to quad (last node must be the 'tip')
const int PromPoly::pyramidID_[5][4]={{4,2,1,-9},{4,3,2,-9},{4,0,3,-9},
				      {4,1,0,-9},{0,1,2,3}};
const int PromPoly::octID_[6][4] ={{0,1,2,3},{5,6,2,1},{7,6,5,4},
				   {3,2,6,7},{0,3,7,4},{4,5,1,0} };
const int PromPoly::septID_[6][4] ={ {0,1,2,3},{0,4,5,1},{0,3,4,-9},
				     {2,6,3,-9},{1,5,6,2},{5,4,3,6}};
const int PromPoly::wedgeID_[5][4] ={ {0,1,2,3},{3,4,0,-9},{2,1,5,-9},
				      {0,4,5,1},{3,2,5,4}};
const int PromPoly::quadID_[4][4]={{0,1,-9,-9},{1,2,-9,-9},
				   {2,3,-9,-9},{3,4,-9,-9}}; 
#if defined(__cplusplus)
extern "C" { 
#endif
  extern double orient3d(double *, double *, double *, double *);
  extern double insphere(double *, double *, double *, double *, double *);
#if defined(__cplusplus)
}
#endif
 
#undef __FUNCT__  
#define __FUNCT__ "PromElement" 
//////////////////////////////////////////////////////////////////////
// PromElement - general mesh object
//
/*  PromElement::PromElement ******************************************
 *  - constructors
 */
PromElement::PromElement() 
{
  reset();
}
/*  PromElement::reset ******************************************
 *  - constructors
 */
void PromElement::reset()
{
  // clear bits
  bits_.state = PROM_UNKNOWN;
  bits_.topo = PROM_INTERIOR;
  bits_.color = 0;
  bits_.type = PROM_NOELEM;
  bits_.exterior = 0;
  bits_.dead = 0; 
  bits_.qual = 0;
  bool deff = FALSE;
  bits_.dof0 = deff;
  bits_.dof1 = deff;
  bits_.dof2 = deff;
  bits_.dof3 = deff;
  bits_.dof4 = deff;
  bits_.dof5 = deff;
  bits_.dof6 = deff;
  bits_.dof7 = deff;
  bits_.dof8 = deff;
  bits_.dof9 = deff;
  bits_.dof10 = deff;
  bits_.dof11 = deff;
  bits_.dof12 = deff;
  bits_.dof13 = deff;
  bits_.dof14 = deff;
  bits_.dof15 = deff;
  bits_.dof16 = deff;
  bits_.dof17 = deff;
  bits_.isenv = FALSE;

  bits_.done = 0;
  bits_.mat_topo = 0;
  bits_.malloced = 0; 
  bits_.pointer = 0; 
  bits_.shp_sz = 0;
  bits_.aux = 0;
  bits_.ndf = 0;
}

// copy op
PromElement &PromElement::operator=( const PromElement& a )
{
  bits_.dof0 = a.bits_.dof0;
  bits_.dof1 = a.bits_.dof1;
  bits_.dof2 = a.bits_.dof2;
  bits_.dof3 = a.bits_.dof3;
  bits_.dof4 = a.bits_.dof4;
  bits_.dof5 = a.bits_.dof5;
  bits_.dof6 = a.bits_.dof6;
  bits_.dof7 = a.bits_.dof7;
  bits_.dof8 = a.bits_.dof8;
  bits_.dof9 = a.bits_.dof9;
  bits_.dof10 = a.bits_.dof10;
  bits_.dof11 = a.bits_.dof11;
  bits_.dof12 = a.bits_.dof12;
  bits_.dof13 = a.bits_.dof13;
  bits_.dof14 = a.bits_.dof14;
  bits_.dof15 = a.bits_.dof15;
  bits_.dof16 = a.bits_.dof16;
  bits_.dof17 = a.bits_.dof17;
  bits_.mat_topo = a.bits_.mat_topo;
  bits_.state = a.bits_.state;
  bits_.topo = a.bits_.topo;
  bits_.type =a.bits_.type;
  bits_.isenv = a.bits_.isenv;  
  bits_.ndf = a.bits_.ndf;

  //bits_.exterior = a.bits_.exterior; 
  //bits_.color = a.bits_.color;
  bits_.qual = a.bits_.qual;
  bits_.aux = a.bits_.aux;         // used by send/recv
  bits_.pointer = a.bits_.pointer; // used by send/recv
  //bits_.shp_sz = a.bits_.shp_sz;
  //bits_.done = a.bits_.done; 
  //bits_.malloced = a.bits_.malloced;  
  //bits_.dead = a.bits_.dead;

  return *this; 
}

/* PromElement::Write/Read ***********************************************
 *
 *   - Read/Write
 *
 *  INPUT:
 *    -  buff: buffer to read/write with
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code.
 */
#undef __FUNCT__  
#define __FUNCT__ "Write/Read" 
int PromElement::Read( int **ppb )
{
  // volitile data not to be copied
  int oldm = bits_.malloced, oldd = bits_.done;
  int olde = bits_.dead,    oldz = bits_.shp_sz;
  if(olde != 0)SETERRQ(PETSC_COMM_SELF,1,"PromElement::Read : bits_.dead");
  struct bitsTAG *pb = (struct bitsTAG*)*ppb; 
  bits_ = *pb++;   
  if(bits_.dead != 0)SETERRQ(PETSC_COMM_SELF,1,"bits_.dead != 0");
  *ppb = (int*)pb;
  bits_.malloced = oldm; bits_.done = oldd;
  bits_.dead = olde;     bits_.shp_sz = oldz;

  return 0;
}

int PromElement::Write( int **ppb )const
{ 
  struct bitsTAG *pbt = (struct bitsTAG*)*ppb; 
  *pbt++ = bits_; 
  *ppb = (int*)pbt;                                   
  return 0;
} 
 
#undef __FUNCT__  
#define __FUNCT__ "setFixed"  
/*  PromElement::setFixed ******************************************
 *  - 
 */
bool PromElement::setFixed ( int tp, bool f /*= TRUE */) 
{ 
  assert(this!=NULL);
  bool r; 
  switch (tp) {
  case 0 : r = bits_.dof0;  bits_.dof0  = f; return r;
  case 1 : r = bits_.dof1;  bits_.dof1  = f; return r;
  case 2 : r = bits_.dof2;  bits_.dof2  = f; return r;
  case 3 : r = bits_.dof3;  bits_.dof3 = f; return r;
  case 4 : r = bits_.dof4;  bits_.dof4 = f; return r;
  case 5 : r = bits_.dof5;  bits_.dof5 = f; return r;
  case 6 : r = bits_.dof6;  bits_.dof6  = f; return r;
  case 7 : r = bits_.dof7;  bits_.dof7  = f; return r;
  case 8 : r = bits_.dof8;  bits_.dof8  = f; return r;
  case 9 : r = bits_.dof9;  bits_.dof9  = f; return r;
  case 10 : r = bits_.dof10;  bits_.dof10 = f; return r;
  case 11 : r = bits_.dof11;  bits_.dof11 = f; return r;
  case 12 : r = bits_.dof12;  bits_.dof12 = f; return r;
  case 13 : r = bits_.dof13;  bits_.dof13  = f; return r;
  case 14 : r = bits_.dof14;  bits_.dof14 = f; return r;
  case 15 : r = bits_.dof15;  bits_.dof15 = f; return r;
  case 16 : r = bits_.dof16;  bits_.dof16 = f; return r;
  case 17 : r = bits_.dof17;  bits_.dof17 = f; return r;
  default : assert(0);
  }
  return FALSE; 
}

/*  PromElement::isFixed ******************************************
 *  - 
 */
bool PromElement::isFixed ( const int ndf, int type /*= -9*/ ) const 
{ 
  assert(this!=NULL);
  int ii;
  switch (type) {
  case 0 : return bits_.dof0;
  case 1 : return bits_.dof1;
  case 2 : return bits_.dof2;
  case 3 : return bits_.dof3;
  case 4 : return bits_.dof4;
  case 5 : return bits_.dof5;
  case 6 : return bits_.dof6;
  case 7 : return bits_.dof7;
  case 8 : return bits_.dof8;
  case 9 : return bits_.dof9;
  case 10 : return bits_.dof10;
  case 11 : return bits_.dof11;
  case 12 : return bits_.dof12;
  case 13 : return bits_.dof13;
  case 14 : return bits_.dof14;
  case 15 : return bits_.dof15;
  case 16 : return bits_.dof16;
  case 17 : return bits_.dof17;
  case -9 :
    for( ii = 0 ; ii < ndf ; ii++ ) 
      if( isFixed(ndf,ii) ) return TRUE;
    return FALSE;
  default: assert(0);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "reset" 
//////////////////////////////////////////////////////////////////////
// PromPoly - general mesh edge
//
// general constructor
void PromPoly::reset( const int numnp, PromNode *a[] )
{
  assert( this!=NULL && numnp == numNodes() );  
  assert(nodes_!=NULL);
  PromElement::reset();

  for ( int i = 0 ; i < numnp ; i++ ){ nodes_[i] = a[i]; }
  for ( int j = 0 ; j < numNeigh(); j++ ) neigh_[j] = NULL;
  bits_.dead = 0;
}

#undef __FUNCT__  
#define __FUNCT__ "reset" 
//////////////////////////////////////////////////////////////////////
// PromDTet - Delaunay tetrahedron
//
void PromDTet::reset()
{
  assert( this!=NULL );  assert(bits_.dead);  PromPoly::reset(); 
  
  bits_.type = PROM_SOLID; //
  quality_ = tdttVol_ = 0.0;
  bits_.qual = 0;
  
  nodes_[0] = NULL;
  nodes_[1] = NULL;
  nodes_[2] = NULL;
  nodes_[3] = NULL;

  neigh_[0] = NULL;
  neigh_[1] = NULL;
  neigh_[2] = NULL;
  neigh_[3] = NULL;
} 
void PromDTet::reset( PromNode *a, PromNode *b, PromNode *c, PromNode *d )
{
  assert( this!=NULL );  PromPoly::reset();
  
  bits_.type = PROM_SOLID; 
  quality_ = tdttVol_ = 0.0;
  bits_.qual = 0;

  nodes_[0] = a;  assert(a!=NULL);
  nodes_[1] = b;  assert(b!=NULL);
  nodes_[2] = c;  assert(c!=NULL);
  nodes_[3] = d;  assert(d!=NULL);
  bits_.dead = 0;
    
  neigh_[0] = NULL;
  neigh_[1] = NULL;
  neigh_[2] = NULL;
  neigh_[3] = NULL;
  bits_.dead = 0;
}
#undef __FUNCT__  
#define __FUNCT__ "operator=" 
/* PromDTet::operator= ********************************************
 *
 */
PromDTet& PromDTet::operator=( const PromDTet& a )
{ 
  PromElement::operator=( a );
  //index = a.index;
  assert(this!=NULL);
  assert ( numNodes() == a.numNodes() );
  
  bits_.dead = a.bits_.dead;     
  bits_.type = a.bits_.type; //
  
  nodes_[0] = a.nodes_[0];  
  nodes_[1] = a.nodes_[1];  
  nodes_[2] = a.nodes_[2];  
  nodes_[3] = a.nodes_[3];  
  
  neigh_[0] = a.neigh_[0];
  neigh_[1] = a.neigh_[1];
  neigh_[2] = a.neigh_[2];
  neigh_[3] = a.neigh_[3];
  
  quality_ = a.quality_;
  
  return *this; 
}
#undef __FUNCT__  
#define __FUNCT__ "willStrandNode" 
/* PromDTet::willStrandNode ********************************************
 *
 *  INPUT:
 *    - ni: number of one element nodes (out)
 *
 */
bool PromDTet::willStrandNode( int &ni )const 
{
  int kk,ii;
  for( kk = 0, ni = 0; kk < 4 ; kk++ ){ 
    PromGeoNode *gc = dynamic_cast<PromGeoNode*>(nodes_[kk]);
    ii = gc->symbData_->elemList.getCount(); assert(ii>0);
    if( ii == 1 ) {
      ni++;
      if( nodes_[kk]->numChildren() ) return TRUE;
    }
  }
  return FALSE;
}

#undef __FUNCT__  
#define __FUNCT__ "calcTestData" 
/* PromDTet::calcTestData ***********************************************
 *
 *   - constructor helper - set cashed data to test 'isInCirc',
 *     evalute 5*5 determindate
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - volume ofxc element
 */
enum {X=0,Y,Z};
gfloat PromDTet::calcTestData()
{
#ifdef PROM_HAVE_PRED
  // node coordinate pointers - permute to getnto positive orientation
  PromGeoNode *nn1 = dynamic_cast<PromGeoNode*>(nodes_[0]);
  PromGeoNode *nn2 = dynamic_cast<PromGeoNode*>(nodes_[1]);
  PromGeoNode *nn3 = dynamic_cast<PromGeoNode*>(nodes_[2]);
  PromGeoNode *nn4 = dynamic_cast<PromGeoNode*>(nodes_[3]);
  PromCoord  *c4 = &nn4->symbGeoData_->coord, *c1 = &nn1->symbGeoData_->coord;
  PromCoord  *c2 = &nn2->symbGeoData_->coord, *c3 = &nn3->symbGeoData_->coord;

  tdttVol_ = -orient3d( c1->data_, c2->data_, c3->data_, c4->data_ ); // /6 ???
#else
  tdttVol_ = 0.;
#endif
  return tdttVol_;
}

#undef __FUNCT__  
#define __FUNCT__ "isInCirc" 
/* PromDTet::isInCirc ***********************************************
 *
 *    - test if coord is in the circumspere of this
 *
 *   INPUT:
 *      - coord: coordinate to test.
 *      - t2: dot product of 'coord' (out)
 *
 *   SIDE EFFECTS:
 *      - 8 flops
 *
 *   RETURN:
 *      - is it in the circumspere of this
 *      - diff, used to recover from close to zero errors
 */
bool PromDTet::isInCirc( PromCoord &cc, double *t2 )
{
#ifdef PROM_HAVE_PRED
  PromGeoNode *nn0 = dynamic_cast<PromGeoNode*>(nodes_[0]);
  PromGeoNode *nn1 = dynamic_cast<PromGeoNode*>(nodes_[1]);
  PromGeoNode *nn2 = dynamic_cast<PromGeoNode*>(nodes_[2]);
  PromGeoNode *nn3 = dynamic_cast<PromGeoNode*>(nodes_[3]);
  PromCoord  *c3 = &nn3->symbGeoData_->coord, *c0 = &nn0->symbGeoData_->coord;
  PromCoord  *c1 = &nn1->symbGeoData_->coord, *c2 = &nn2->symbGeoData_->coord;

  *t2 = insphere( c3->data_, c2->data_, c1->data_, c0->data_, cc.data_ );
#else
  *t2 = 0.;
#endif
  return ( *t2 < 0.0 );
}

#undef __FUNCT__  
#define __FUNCT__ "isIn" 
/* PromDTet::isIn() ***********************************************
 *
 *    - return 1 if coordinate is in tetrahedra of this, 0 - if not
 *
 *  INPUT:
 *    - crd: coordinate to check if it is in here (must not be on top)
 *    - shp: array to put shape fuction values (out)
 *    - tol: tloerance to use in check
 *    - ok: return flag of is solve ok (i.e. is 'shp' valid) (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - if it is in this
 */
bool PromDTet::isIn( const PromCoord &crd, shpfloat shp[4], const gfloat tol,
		     int mype, bool *ok ) 
{
  *ok = shape( nodes_, crd, shp ); 
  if( !*ok ) {
    shp[0] = shp[1] = shp[2] = shp[3] = 0.0;
    return FALSE;
  }
  else {
    return (shp[0] >= -tol && shp[1] >= -tol && 
	    shp[2] >= -tol && shp[3] >= -tol);
  }
}

#undef __FUNCT__  
#define __FUNCT__ "SetNeigh"
/* PromDTet::SetNeigh ***********************************************
 *
 *    - set 'neigh' with new element 
 *
 *   INPUT:n
 *      - newel: new elem (could be NULL)
 *      - n1, n2, n3: nodes of face to set
 *
 *   SIDE EFFECTS:
 *      - sets 'neigh'
 *
 *   RETURN:
 *      - index in elem that is set, -9 for not found 
 */
int PromDTet::SetNeigh(PromDTet *newel, const PromNode *n1, const PromNode *n2,
		    const PromNode *n3, const PromNode *n4 /* = NULL */)
{
  // for all faces of outside 'elem'
  const int numadj1 = numNeigh();
  const int *elemID = getIDArr();
  const int NUM1 = (!n4) ? 3 : 4;
  for ( int ii = 0 ; ii < numadj1 ; ii++ ) {            
    const int NUM2 = elemID[ ii * 4 + 3 ] == -9 ? 3 : 4;    
    if ( NUM2 != NUM1 ) continue;
    
    // for all nodes of face            
    int found = 0;
    for ( int jj = 0 ; jj < NUM2 ; jj++ ) {
      
      PromNode *curr = nodes_[  elemID[ ii * 4 + jj ] ]; assert(curr!=NULL);
      if ( curr == n1 || curr == n2 || curr == n3 || curr == n4 )
        found++;
    }
    if ( found == NUM1 ) {
      return ii;
    }
  }
  
  return -9;
}
#undef __FUNCT__  
#define __FUNCT__ "isSliver" 
/* PromDTet::isSliver() *************************************
 * 
 */
bool  PromDTet::isSliver()
{ 
  assert(this!=NULL);  
  return ( quality() < 1.e-8 );
} 

#undef __FUNCT__  
#define __FUNCT__ "calcQuality" 
/* PromDTet::calcQuality() *************************************
 *
 *  - quality metric for a tet.
 *
 *  INPUT:
 *
 *  SIDE FFECTS:
 *
 *  - set 'quality' ala Golias and Tsiboukis (12/2^{1/2} * volume/ Lmax^3)
 *  - sets guesse of largest surface, to swap on in 'bits_.side'
 *    and nodes of longest side in 'bits_.n1' and 'bits_.n2' 
 *
 *  RETURN:
 *  - quality value
 */
shpfloat PromDTet::calcQuality()
{
  assert( numNodes() == 4 );
  PromGeoNode *nn0 = dynamic_cast<PromGeoNode*>(nodes_[0]);
  PromGeoNode *nn1 = dynamic_cast<PromGeoNode*>(nodes_[1]);
  PromGeoNode *nn2 = dynamic_cast<PromGeoNode*>(nodes_[2]);
  PromGeoNode *nn3 = dynamic_cast<PromGeoNode*>(nodes_[3]);

  PromCoord &fp1 = nn0->symbGeoData_->coord, &fp2 = nn1->symbGeoData_->coord;
  PromCoord &fp3 = nn2->symbGeoData_->coord, &fp4 = nn3->symbGeoData_->coord;
  
  // array of vectors for sides
  double vec[6][3] = {
    { fp2[0] - fp1[0], fp2[1] - fp1[1], fp2[2] - fp1[2]},
    { fp3[0] - fp1[0], fp3[1] - fp1[1], fp3[2] - fp1[2]},
    { fp4[0] - fp1[0], fp4[1] - fp1[1], fp4[2] - fp1[2]},
    { fp3[0] - fp2[0], fp3[1] - fp2[1], fp3[2] - fp2[2]},
    { fp4[0] - fp2[0], fp4[1] - fp2[1], fp4[2] - fp2[2]},
    { fp4[0] - fp3[0], fp4[1] - fp3[1], fp4[2] - fp3[2]} };

  // get first two longest sides
  double dot, max1 = 0.0, max2 = 0.0;
  for ( int i = 0 ;i < 6 ; i++ ) {
    dot = vec[i][0] * vec[i][0] +
      vec[i][1] * vec[i][1] + vec[i][2] * vec[i][2];
    if ( dot > max1 ) {
      if ( max1 > max2 ) max2 = max1;
      max1 = dot;
    }
    else if ( dot > max2 ) max2 = dot;
  }
  double len = sqrt(max1);

  // heuristic to pick longest side - if == 6 then no good for swap anyway
  // vol = 6 * volume
  double vol = tdttVol_;
  const double twelve_sqrt2 = 8.48528; dot = len * max1; assert(dot>0.0);
  quality_ = twelve_sqrt2 * vol / dot;

  return quality_;
}

#undef __FUNCT__  
#define __FUNCT__ "shape"   
/* PromDTet::shape() ***********************************************
 *
 *   - static - sets linear triangle shape function values
 * 
 *  INPUT:
 *    - nodes: array of node pointers (i.e. CElement::nodes)
 *    - crd: coordinate of point
 *    - L: shape functiona values of this at 'coord'
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - is element OK 
 */  
bool PromDTet::shape( PromNode *nnds[], const PromCoord &cc, shpfloat L[])
{
  assert(nnds && L);
  double   a[4][5], temp, ajj; //double
  PromGeoNode *nn0 = dynamic_cast<PromGeoNode*>(nnds[0]);
  PromGeoNode *nn1 = dynamic_cast<PromGeoNode*>(nnds[1]);
  PromGeoNode *nn2 = dynamic_cast<PromGeoNode*>(nnds[2]);
  PromGeoNode *nn3 = dynamic_cast<PromGeoNode*>(nnds[3]);
  PromCoord  &c3 = nn3->symbGeoData_->coord, &c0 = nn0->symbGeoData_->coord;
  PromCoord  &c1 = nn1->symbGeoData_->coord, &c2 = nn2->symbGeoData_->coord;

  // set matrix to solve, with rhs in ndm+1 column
  a[3][0] = 1.0;
  a[3][1] = 1.0;
  a[3][2] = 1.0;
  a[3][3] = 1.0;
  
  a[0][0] = c0[0]; 
  a[1][0] = c0[1]; 
  a[2][0] = c0[2]; 
  a[0][4] = cc[0]; // rhs
  
  a[0][1] = c1[0]; 
  a[1][1] = c1[1]; 
  a[2][1] = c1[2]; 
  a[1][4] = cc[1]; // rhs
  
  a[0][2] = c2[0]; 
  a[1][2] = c2[1]; 
  a[2][2] = c2[2]; 
  a[2][4] = cc[2]; // rhs
  
  a[0][3] = c3[0]; 
  a[1][3] = c3[1]; 
  a[2][3] = c3[2]; 
  a[3][4] = 1.0;
  
  // forward elimination - row 2-4
  PromContext::pivot45( 0, a );
  ajj = a[0][0];           if( fabs(ajj) < 1.e-100 ) return FALSE; 
  temp = a[1][0] / ajj;    
  a[1][1] -= a[0][1]*temp;
  a[1][2] -= a[0][2]*temp; 
  a[1][3] -= a[0][3]*temp; 
  a[1][4] -= a[0][4]*temp;
  
  temp = a[2][0] / ajj;    
  a[2][1] -= a[0][1]*temp;
  a[2][2] -= a[0][2]*temp; 
  a[2][3] -= a[0][3]*temp; 
  a[2][4] -= a[0][4]*temp;
  
  temp = a[3][0] / ajj;    
  a[3][1] -= a[0][1]*temp;
  a[3][2] -= a[0][2]*temp; 
  a[3][3] -= a[0][3]*temp; 
  a[3][4] -= a[0][4]*temp;
  
  // forward elimination - row 3-4
  PromContext::pivot45(1, a);
  ajj = a[1][1];          if( fabs(ajj) < 1.e-100 ) return FALSE; 
  temp = a[2][1] / ajj;    
  a[2][2] -= a[1][2]*temp; 
  a[2][3] -= a[1][3]*temp; 
  a[2][4] -= a[1][4]*temp;
  
  temp = a[3][1] / ajj;    
  a[3][2] -= a[1][2]*temp; 
  a[3][3] -= a[1][3]*temp; 
  a[3][4] -= a[1][4]*temp;
  
  // forward elimination - row 4
  PromContext::pivot45( 2, a );
  ajj = a[2][2];           if( fabs(ajj) < 1.e-100 ) return FALSE; 
  temp = a[3][2] / ajj;    
  a[3][3] -= a[2][3]*temp; 
  a[3][4] -= a[2][4]*temp;
  
  // back sub 
  ajj = a[3][3];           if( fabs(ajj) < 1.e-100 ) return FALSE; 
  temp = a[3][4] / ajj;
  a[2][4] -= a[2][3] * temp;
  a[1][4] -= a[1][3] * temp;
  a[0][4] -= a[0][3] * temp;
  L[3] = temp;  //copy answer
  
  temp = a[2][4] / a[2][2];
  a[1][4] -= a[1][2] * temp;
  a[0][4] -= a[0][2] * temp;
  L[2] = temp;   //copy answer
  
  temp = a[1][4] / a[1][1];
  a[0][4] -= a[0][1] * temp;
  L[1] = temp;   //copy answer
  
  L[0] = a[0][4] / a[0][0];  //copy answer 
  // check for OK
  double tt = fabs( L[0] + L[1] + L[2] + L[3] - 1.0 );
  if( tt > 0.0001 ) return FALSE;
  
  return TRUE;
}

#undef __FUNCT__  
#define __FUNCT__ "" 
/*  PromFacet::reset ******************************************
 *
 *  - constructors
 */
void PromFacet::reset()
{
  assert(this!=NULL); PromPoly::reset();
  face_ = 0;
  norm_[0]=0.0;
  norm_[1]=0.0;
  norm_[2]=0.0;
  
  nodes_[0]=NULL;
  nodes_[1]=NULL;
  nodes_[2]=NULL;
  nodes_[3]=NULL;
  
  bits_.dead = 1;
  bits_.done = 0;
  bits_.exterior = 0;

  assert(isChildListEmpty());
}
#undef __FUNCT__  
#define __FUNCT__ "reset" 
/*  PromFacet::reset ******************************************
 *
 *  - constructors
 *     - marks faces that are bad (dup. nodes)
 */
void PromFacet::reset( PromNode *n1, PromNode *n2, PromNode *n3, PromNode *n4 )
{
  assert( this && n1 && n2 && n3 );
  face_ = 0;
  bits_.dead = 0;
  bits_.done = 0;
  bits_.exterior = 0;
 
  // watch for dups  
  //if ( n1 == n2 ) { n2 = n3; n3 = n4;}
  //if ( n2 == n3 || n1 == n3 ) n3 = n4;
  //if ( n1 == n4 || n2 == n4 || n3 == n4 ) n4 = NULL;
  assert( n1 != n2 && n2 != n3 && n3 != n1 );
  PromCoord &c3 = dynamic_cast<PromGeoNode*>(n3)->symbGeoData_->coord;
  PromCoord &c1 = dynamic_cast<PromGeoNode*>(n1)->symbGeoData_->coord;
  PromCoord &c2 = dynamic_cast<PromGeoNode*>(n2)->symbGeoData_->coord;
  PromContext::cross( c1, c2, c3, &norm_ ); // set norm
  if( !n4 ) norm_ *= 0.5; // triagle has half area
  bits_.done = 0; 

  nodes_[0] = n1;
  nodes_[1] = n2;
  nodes_[2] = n3;
  nodes_[3] = n4;

  assert(isChildListEmpty());
}
 
#undef __FUNCT__  
#define __FUNCT__ "PromNPoly" 
/* PromNPoly::PromNPoly ******************************************
 *
 */
PromNPoly::PromNPoly( PromNode **pn, const int nn )
{ 
  SetNumNodes(nn);
  neigh_ = NULL;
  for( int i = 0 ; i < nn ; i++ ) { nodes_[i] = pn[i]; }
  setType( PROM_N_POLY );
  bits_.dead = 0;
}

#undef __FUNCT__  
#define __FUNCT__ "~PromNPoly" 
/* PromNPoly::~PromNPoly ******************************************
 *
 */
PromNPoly::~PromNPoly()
{
  assert(this!=NULL);assert( nodes_private_ != NULL );
  PetscFree( nodes_private_ );
}

#undef __FUNCT__  
#define __FUNCT__ "numNeigh" 
/* PromNPoly::numNeigh ******************************************
 *
 */
int PromNPoly::numNeigh() const
{
  //SETERRQ(PETSC_COMM_SELF,1,"numNeigh not supported for PromNPoly");
  return 0;
}

int PromNPoly::SetNumNodes( const int nn )
{
  assert(this!=NULL); assert(nn>0); assert(bits_.shp_sz == 0); // init only!!!
  if( bits_.shp_sz != (unsigned)nn ) {
    if( nodes_ != NULL ) PetscFree( nodes_ );
    int ierr = PetscMalloc(nn*sizeof(PromNode*),&nodes_); CHKERRQ(ierr);
    nodes_private_ = nodes_; 
    assert(nn < PROM_MAX_SHP_SZ);
    bits_.shp_sz = nn;
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "Archive" 
/* PromElement::Archive() ***********************************************
 *
 *   - Save/Write to file 
 *
 *  INPUT:
 *    -  file: file to read/write
 *    -  type: READ, WRITE, PRINT(pretty debug output)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code.
 */
int PromElement::Archive( FILE *file, const PromArchiveType type )
{
  int i; 
  char argstr[] = "%c%c%c%c%c%c%c%c";
  
  if( type == PROM_PRINT ) {
    //fprintf( file, "PromElement:%d ",bits_);
  }
  else if( type == PROM_WRITE ) {
    char chars[PROM_MAXNDF] = 
      { (char)bits_.dof0, (char)bits_.dof1, (char)bits_.dof2, 
	(char)bits_.dof3, (char)bits_.dof4, (char)bits_.dof5,
	(char)bits_.dof6, (char)bits_.dof7 };
    fprintf( file, argstr, chars[0], chars[1], chars[2], 
	     chars[3], chars[4], chars[5] , chars[6], chars[7] );
    fprintf( file, " %d",bits_.shp_sz);
  }
  else { 
    char chars[PROM_MAXNDF];
    fscanf( file, argstr, &chars[0], &chars[1], &chars[2], &chars[3], 
	    &chars[4], &chars[5], &chars[6], &chars[7]  ); 
    bits_.dof0 = chars[0]; bits_.dof1 = chars[1]; bits_.dof2 = chars[2];
    bits_.dof3 = chars[3]; bits_.dof4 = chars[4]; bits_.dof5 = chars[5];
    bits_.dof6 = chars[6]; bits_.dof7 = chars[7];
    fscanf( file," %d",&i );  bits_.shp_sz = i; 
  }
  
  return 0;
}
